from bisect import bisect
from typing import Tuple, Union, List, Optional
from transparentpath import TransparentPath

import numpy as np
import pandas as pd

from .configs import NodeDataConfig
from .loader import load_y
import logging
logger = logging.getLogger(__name__)


class DataPrep:
    """Abstract class for DataPrep."""

    def __init__(self, data: NodeDataConfig, **kwargs):
        """
        Parameters
        ----------
        data: NodeDataConfig
            `ifra.node.Node`'s *data*
        kwargs:
            Any additionnal keyword argument that the overleading class accepts. Those arguments will become attributes.
        """
        self.data = data
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])

    def dataprep(self):
        """Writes the output of the dataprep in `ifra.node.Node`'s *data.x_path* and `ifra.node.Node`'s *data.y_path*
        parent directories by appending *_datapreped* to the files names. Modifies `ifra.node.Node`'s
        *data.x_datapreped_path* and `ifra.node.Node`'s *data.y_datapreped_path* to point to those files.
        """
        x, y = self.dataprep_method(
            self.data.x_path.read(**self.data.x_read_kwargs),
            load_y(self.data.y_path, **self.data.y_read_kwargs)
        )
        x_suffix = self.data.x_path.suffix
        y_suffix = self.data.y_path.suffix
        self.data.x_datapreped_path = self.data.x_path.with_suffix("").append("_datapreped").with_suffix(x_suffix)
        self.data.y_datapreped_path = self.data.y_path.with_suffix("").append("_datapreped").with_suffix(y_suffix)

        self.data.x_datapreped_path.write(x)
        self.data.y_datapreped_path.write(y)

    def dataprep_method(self, x: pd.DataFrame, y: pd.Series) -> Tuple[pd.DataFrame, pd.Series]:
        """To be implemented in daughter class.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            Datapreped x and y
        """
        pass


class BinFeaturesDataPrep(DataPrep):
    """Overloads DataPrep class. Bins each feature columns in 'nbins' modalities. Does not modify y.

    Can be used by giving *binfeatures* as *dataprep* configuration when creating a `ifra.node.Node`

    Attributes
    ----------
    nbins: int
        Number of bins to use
    """

    def __init__(self, data: NodeDataConfig, nbins: int, bins: dict, save_bins: Optional[TransparentPath]):
        if save_bins is not None:
            save_bins = TransparentPath(save_bins)
        super().__init__(data, nbins=nbins, bins=bins, save_bins=save_bins)

    def dataprep_method(self, x, y):

        # noinspection PyUnresolvedReferences
        def to_apply(xx: pd.Series):
            if xx.name in self.bins:
                bins = self.bins[xx.name]
            else:
                bins = get_bins(xx, self.nbins)
                self.bins[xx.name] = bins

            to_ret = discretize(xx, bins)

            if self.save_bins is not None:
                if not self.save_bins.isfile():
                    self.save_bins.write({xx.name: bins})
                else:
                    allbins = self.save_bins.read()
                    allbins[xx.name] = bins
                    self.save_bins.write(allbins)
            return to_ret

        def find_bins(xx: pd.Series, nbins: int) -> np.ndarray:
            """
            Function used to find the bins to discretize xcol in nbins modalities

            Parameters
            ----------
            xx : pd.Series
               Series to discretize

            nbins: int
                number of modalities

            Return
            ------
            bins: np.ndarray
               the bins for disretization (result from numpy percentile function)
            """
            q_list = np.arange(100.0 / nbins, 100.0, 100.0 / nbins)
            bins = np.array([np.nanpercentile(xx, i) for i in q_list])
            return bins

        def get_bins(xx: pd.Series, nb_bucket: int) -> np.ndarray:
            logger.debug(f"Getting bins for {xx.name}...")
            if nb_bucket == 0:
                raise ValueError("nb_bucket must be greater than 0")
            if nb_bucket == 1:
                return np.array([])
            if len(xx) == 0 or len(xx) == 1:
                return np.array([])
            if len(np.unique(xx)) <= nb_bucket:
                return np.ediff1d(np.unique(xx)) / 2. + np.unique(xx)[:-1]
            bins = find_bins(xx, nb_bucket)
            while len(set(bins.round(5))) != len(bins):
                nb_bucket -= 1
                bins = find_bins(xx, nb_bucket)
            if len(bins) != nb_bucket - 1:
                raise ValueError(f"Error in get_bins : {len(bins) + 1} bins where found but {nb_bucket} were asked.")
            logger.debug(f"... got bins for {xx.name}")
            return bins

        def discretize(xx: pd.Series, bins: Union[np.ndarray, List[float]]) -> pd.Series:
            """
            Transform a Series of float to a Series if int, where each float is replaced by the bin it matches.

            Parameters
            ----------
            xx : pd.Series
                Series to discretize
            bins : Union[np.ndarray, List[float]]
                The list of bins in the form e.g. [0.1, 0.3, ... 0.8] :
                    // -inf ---> 0.1 is bin 0
                    // 0.1 ---> 0.3 is bin 1, etc.
                    // 0.8 ---> +inf is the last bin

            Return
            ------
            discrete_x : pd.Series
                The discretization of x

            """
            logger.debug(f"Discretizing {xx.name}...")
            mask = np.isnan(xx)
            discrete_x = xx.apply(lambda var: bisect(bins, var))
            discrete_x[mask] = np.nan
            logger.debug(f"... discretized {xx.name}")
            return discrete_x

        # noinspection PyUnresolvedReferences
        x = x.apply(lambda xx: to_apply(xx), axis=0)
        return x, y



from ifra.central_server import CentralServer
from ifra.configs import CentralConfig, Config


def test_simple_init_and_run():
    server = CentralServer(central_configs=CentralConfig("tests/data/central_configs.json"))
    assert isinstance(server.central_configs, (Config, CentralConfig))

    assert server.emitter.doing is None
    assert server.emitter.error is None

    server.run(timeout=1, sleeptime=0.1)

    assert server.emitter.doing is None
    assert server.emitter.error is None

    server = CentralServer(central_configs=CentralConfig("tests/data/central_configs.json"))
    assert isinstance(server.central_configs, (Config, CentralConfig))
    server.run(timeout=1, sleeptime=0.1)



import pandas as pd

from ifra import Node
from ifra.configs import Config, NodeLearningConfig, NodeDataConfig
from ifra.fitters import Fitter, DecisionTreeClassificationFitter
from ifra.node_model_updaters import NodeModelUpdater, AdaBoostNodeModelUpdater
from ifra.datapreps import DataPrep, BinFeaturesDataPrep
from ..alternates.dataprep import AlternateDataPrep
from ..alternates.updater import AlternateUpdater
from ..alternates.fitter import AlternateFitter


def test_init_and_run_simple(clean):
    node = Node(
        learning_configs=NodeLearningConfig("tests/data/node_test/learning_configs.json"),
        data=NodeDataConfig("tests/data/node_test/data_configs.json"),
    )
    assert isinstance(node.learning_configs, (Config, NodeLearningConfig))
    assert isinstance(node.data, (Config, NodeDataConfig))
    assert isinstance(node.fitter, (Fitter, DecisionTreeClassificationFitter))
    assert isinstance(node.updater, (NodeModelUpdater, AdaBoostNodeModelUpdater))
    assert isinstance(node.dataprep, (DataPrep, BinFeaturesDataPrep))
    assert node.datapreped is False
    assert node.data.dataprep_kwargs == node.learning_configs.dataprep_kwargs
    assert node.splitted is False
    assert node.model is None
    assert node.last_fetch is None
    assert not node.learning_configs.node_models_path.is_file()

    assert not (node.data.x_path.parent / "x_datapreped.csv").is_file()
    assert not (node.data.y_path.parent / "y_datapreped.csv").is_file()
    assert not (node.data.x_path.parent / "x_train_0.csv").is_file()
    assert not (node.data.y_path.parent / "y_train_0.csv").is_file()
    assert not (node.data.x_path.parent / "x_test.csv").is_file()
    assert not (node.data.y_path.parent / "y_test.csv").is_file()
    assert not (node.data.x_path.parent / "plots").is_dir()
    assert not (node.data.x_path.parent / "plots_datapreped").is_dir()
    assert not (node.data.x_path.parent / "plots_train_0").is_dir()

    assert node.emitter.doing is None
    assert node.emitter.error is None

    node.run(timeout=1, sleeptime=0.1)

    assert node.splitted
    assert node.datapreped
    assert node.data.x_datapreped_path.is_file()
    assert node.data.x_datapreped_path == node.data.x_path.parent / "x_datapreped.csv"
    assert node.data.y_datapreped_path.is_file()
    assert node.data.y_datapreped_path == node.data.y_path.parent / "y_datapreped.csv"
    assert node.data.x_train_path.is_file()
    assert node.data.x_train_path == node.data.x_path.parent / "x_train_0.csv"
    assert node.data.y_train_path.is_file()
    assert node.data.y_train_path == node.data.y_path.parent / "y_train_0.csv"
    assert node.data.x_train_path == node.data.x_test_path
    assert node.data.y_train_path == node.data.y_test_path
    assert (node.data.x_path.parent / "plots").is_dir()
    assert (node.data.x_path.parent / "bins.json").is_file()
    assert (node.data.x_path.parent / "plots_datapreped").is_dir()
    assert (node.data.x_path.parent / "plots_train_0").is_dir()
    assert (node.learning_configs.node_models_path / f"model_main_{node.filenumber}.csv").is_file()
    assert (node.learning_configs.node_models_path / f"model_{node.filenumber}_0.csv").is_file()

    try:
        pd.testing.assert_frame_equal(node.data.x_path.read(), node.data.x_datapreped_path.read())
    except AssertionError:
        pass
    else:
        raise AssertionError("x and x_datapreped DataFrames should be different")

    pd.testing.assert_frame_equal(node.data.y_path.read(), node.data.y_datapreped_path.read())

    assert node.data.x_datapreped_path.read().values.dtype == int
    assert node.data.x_train_path.read().values.dtype == int
    assert node.data.x_path.read().values.dtype != int

    assert node.emitter.doing is None
    assert node.emitter.error is None


def test_init_alternate_dataprep_updater_fitter(clean):
    node = Node(
        learning_configs=NodeLearningConfig("tests/data/node_test/alternate_learning_configs.json"),
        data=NodeDataConfig("tests/data/node_test/data_configs.json"),
    )
    assert isinstance(node.dataprep, AlternateDataPrep)
    assert isinstance(node.updater, AlternateUpdater)
    assert isinstance(node.fitter, AlternateFitter)

    node.run(timeout=1, sleeptime=0.1)

    assert node.data.x_datapreped_path.read().values.dtype != int
    assert node.data.x_train_path.read().values.dtype != int
    assert node.data.x_path.read().values.dtype != int


def test_init_and_run_central_model_present(clean):
    node = Node(
        learning_configs=NodeLearningConfig("tests/data/node_test_with_central_model/learning_configs.json"),
        data=NodeDataConfig("tests/data/node_test_with_central_model/data_configs.json"),
    )
    node.run(timeout=1, sleeptime=0.1)

    x = node.data.x_path.read()
    x_train = node.data.x_train_path.read()
    assert len(x.index) > len(x_train.index)



from pytest import fixture
from transparentpath import Path


@fixture()
def clean():
    output_root = Path("tests/outputs", fs="local")
    if not output_root.is_dir():
        output_root.mkdir()

    output_dirs = [
        output_root / "node_0",
        output_root / "node_1",
        output_root / "node_2",
        output_root / "node_test",
        output_root / "node_models",
        output_root / "central_server",
        output_root / "aggregator",
    ]
    for adir in output_dirs:
        if not adir.is_dir():
            adir.mkdir()
        else:
            for f in adir.ls():
                if "message" not in str(f):
                    f.rm(ignore_kind=True)
    for f in output_root.glob("model_*"):
        f.rm()

    data_root = Path("tests/data", fs="local")
    data_dirs = [
        data_root / "node_0",
        data_root / "node_1",
        data_root / "node_2",
        data_root / "node_test",
    ]
    for data_dir in data_dirs:
        (data_dir / "plots").rmdir(absent="ignore")
        (data_dir / "plots_datapreped").rmdir(absent="ignore")
        (data_dir / "plots_train_0").rmdir(absent="ignore")
        (data_dir / "plots_test_0").rmdir(absent="ignore")

    (data_root / "node_test" / "x_datapreped.csv").rm(absent="ignore")
    (data_root / "node_test" / "y_datapreped.csv").rm(absent="ignore")
    (data_root / "node_test" / "x_train_0.csv").rm(absent="ignore")
    (data_root / "node_test" / "y_train_0.csv").rm(absent="ignore")
    (data_root / "node_test" / "x_test_0.csv").rm(absent="ignore")
    (data_root / "node_test" / "y_test_0.csv").rm(absent="ignore")

    yield

    (data_root / "node_test" / "x_datapreped.csv").rm(absent="ignore")
    (data_root / "node_test" / "y_datapreped.csv").rm(absent="ignore")
    (data_root / "node_test" / "x_train_0.csv").rm(absent="ignore")
    (data_root / "node_test" / "y_train_0.csv").rm(absent="ignore")
    (data_root / "node_test" / "x_test_0.csv").rm(absent="ignore")
    (data_root / "node_test" / "y_test_0_0.csv").rm(absent="ignore")
    (data_root / "node_test" / "plots").rmdir(absent="ignore")
    (data_root / "node_test" / "plots_datapreped").rmdir(absent="ignore")
    (data_root / "node_test" / "plots_train_0").rmdir(absent="ignore")
    (data_root / "node_test" / "plots_test_0_0").rmdir(absent="ignore")
    (output_root / "node_test").rmdir(absent="ignore")

    (data_root / "node_test_with_central_model" / "x_datapreped.csv").rm(absent="ignore")
    (data_root / "node_test_with_central_model" / "y_datapreped.csv").rm(absent="ignore")
    (data_root / "node_test_with_central_model" / "x_train_0.csv").rm(absent="ignore")
    (data_root / "node_test_with_central_model" / "y_train._0csv").rm(absent="ignore")
    (data_root / "node_test_with_central_model" / "x_test_0.csv").rm(absent="ignore")
    (data_root / "node_test_with_central_model" / "y_test_0.csv").rm(absent="ignore")
    (data_root / "node_test_with_central_model" / "plots").rmdir(absent="ignore")
    (data_root / "node_test_with_central_model" / "plots_datapreped").rmdir(absent="ignore")
    (data_root / "node_test_with_central_model" / "plots_train_0").rmdir(absent="ignore")
    (data_root / "node_test_with_central_model" / "plots_test_0").rmdir(absent="ignore")
    (output_root / "node_test_with_central_model").rmdir(absent="ignore")



import pandas as pd

n_nodes = 3
x = pd.read_csv("x.csv", header=None)
print("min: ", x.min())
print("max: ", x.max())
y = pd.read_csv("y.csv", header=None)

n_observations = len(y.index)

print("observations:", n_observations)

for i in range(n_nodes):
    subx = x.sample(frac=1 / (n_nodes - i))
    index = subx.index
    subx = subx.reset_index(drop=True)
    x = x.drop(index)

    suby = y.loc[index]
    suby = suby.reset_index(drop=True)

    subx.to_csv(f"node_{i}/x.csv")
    suby.to_csv(f"node_{i}/y.csv")



import pandas as pd
from ruleskit import RuleSet
from typing import Tuple
from ifra.node_model_updaters import NodeModelUpdater


class AlternateUpdater(NodeModelUpdater):
    @staticmethod
    def make_update(x: pd.DataFrame, y: pd.Series, ruleset: RuleSet) -> Tuple[pd.DataFrame, pd.DataFrame]:
        ruleset.calc_activation(x.values)
        if ruleset.activation is not None:
            x = x.loc[ruleset.activation == 1]
            y = y[ruleset.activation == 1]
        return x, y



from copy import deepcopy

import pytest

from ifra.aggregator import Aggregator, NodeGate
from ifra.configs import AggregatorConfig, Config
from ruleskit import RuleSet, ClassificationRule, HyperrectangleCondition, RegressionRule
from ifra.duplicated_rules_updater import update_duplicated_rules


predictions_float = [1, 1, 4, 2, 2, 2.1, 2.1, 3]
criterions = [1, 1, 2, 1, 1, 2, 3, 2]
predictions_str = ["a", "a", "aa", "b", "b", "bb", "bb", "c"]

classif_ruleset_float = RuleSet(
    [
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])
        ),
    ]
)

for i in range(len(classif_ruleset_float)):
    # noinspection PyProtectedMember
    classif_ruleset_float[i]._prediction = predictions_float[i]

regress_ruleset = RuleSet(
    [
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])),
    ]
)

for i in range(len(regress_ruleset)):
    # noinspection PyProtectedMember
    regress_ruleset[i]._prediction = predictions_float[i]
    # noinspection PyProtectedMember
    regress_ruleset[i]._criterion = criterions[i]

classif_ruleset_str = RuleSet(
    [
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])
        ),
    ]
)

for i in range(len(classif_ruleset_str)):
    # noinspection PyProtectedMember
    classif_ruleset_str[i]._prediction = predictions_str[i]

classif_expected_ruleset_float = RuleSet(
    [
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])
        ),
    ]
)

regress_expected_ruleset = RuleSet(
    [
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])),
    ]
)

regress_expected_ruleset_weighted = RuleSet(
    [
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])),
    ]
)

regress_expected_ruleset_weighted_reversed = RuleSet(
    [
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 2])),
        RegressionRule(condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])),
    ]
)

classif_expected_ruleset_str = RuleSet(
    [
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 1])
        ),
        ClassificationRule(
            condition=HyperrectangleCondition(features_indexes=[0, 1, 2], bmins=[0, 0, 0], bmaxs=[1, 1, 3])
        ),
    ]
)

# noinspection PyProtectedMember
classif_expected_ruleset_float[0]._prediction = 1
# noinspection PyProtectedMember
classif_expected_ruleset_float[1]._prediction = 3

# noinspection PyProtectedMember
regress_expected_ruleset[0]._prediction = 2
# noinspection PyProtectedMember
regress_expected_ruleset[1]._prediction = 2.05
# noinspection PyProtectedMember
regress_expected_ruleset[2]._prediction = 3

# noinspection PyProtectedMember
regress_expected_ruleset_weighted[0]._prediction = 2.5
# noinspection PyProtectedMember
regress_expected_ruleset_weighted[1]._prediction = 14.5 / 7.
# noinspection PyProtectedMember
regress_expected_ruleset_weighted[2]._prediction = 3

# noinspection PyProtectedMember
regress_expected_ruleset_weighted_reversed[0]._prediction = 1.75
# noinspection PyProtectedMember
regress_expected_ruleset_weighted_reversed[1]._prediction = 2.042857143
# noinspection PyProtectedMember
regress_expected_ruleset_weighted_reversed[2]._prediction = 3

# noinspection PyProtectedMember
classif_expected_ruleset_str[0]._prediction = "a"
# noinspection PyProtectedMember
classif_expected_ruleset_str[1]._prediction = "c"


def one_node(node):
    assert isinstance(node, NodeGate)
    assert node.model is None
    assert node.last_fetch is None
    assert node.ok is True


def test_simple_init_and_watch():
    aggr = Aggregator(aggregator_configs=AggregatorConfig("tests/data/aggregator_configs.json"))
    assert isinstance(aggr.aggregator_configs, (Config, AggregatorConfig))
    assert len(aggr.nodes) == 0  # Empty until aggr.run() is called

    assert aggr.emitter.doing is None
    assert aggr.emitter.error is None

    for node in aggr.nodes:
        one_node(node)

    aggr.run(timeout=1, sleeptime=0.5)

    assert aggr.emitter.doing is None
    assert aggr.emitter.error is None

    aggr = Aggregator(aggregator_configs=AggregatorConfig("tests/data/aggregator_configs.json"))
    assert isinstance(aggr.aggregator_configs, (Config, AggregatorConfig))
    assert len(aggr.nodes) == 0
    for node in aggr.nodes:
        one_node(node)

    aggr.run(timeout=1, sleeptime=0.5)


@pytest.mark.parametrize(
    "ruleset, expected_ruleset, weight, best",
    (
        (classif_ruleset_str, classif_expected_ruleset_str, "equi", "max"),
        (classif_ruleset_float, classif_expected_ruleset_float, "equi", "max"),
        (regress_ruleset, regress_expected_ruleset, "equi", "max"),
        (regress_ruleset, regress_expected_ruleset_weighted, "criterion", "max"),
        (regress_ruleset, regress_expected_ruleset_weighted_reversed, "criterion", "min"),
        (regress_ruleset, regress_expected_ruleset_weighted_reversed, "1 * rule.criterion", "min"),
        (regress_ruleset, regress_expected_ruleset_weighted_reversed, "1 * (rule.criterion)", "min"),
    ),
)
def test_classif_pred_aggr(ruleset, expected_ruleset, weight, best):
    ruleset = deepcopy(ruleset)
    update_duplicated_rules(ruleset, weight, best)

    assert len(ruleset) == len(expected_ruleset)
    for rule, exprule in zip(ruleset, expected_ruleset):
        assert rule == exprule
        assert rule.condition == exprule.condition
        if isinstance(rule.prediction, str):
            assert rule.prediction == exprule.prediction
        else:
            assert round(rule.prediction, 6) == round(exprule.prediction, 6)



from time import sleep

from ifra import CentralServer, Node, Aggregator, CentralConfig, NodeDataConfig, NodeLearningConfig, AggregatorConfig
from multiprocessing import Process
import os

max_cpus = min(4, os.cpu_count())

nodes_learning_config = [
    NodeLearningConfig("tests/data/node_0/learning_configs.json"),
    NodeLearningConfig("tests/data/node_1/learning_configs.json"),
    NodeLearningConfig("tests/data/node_2/learning_configs.json"),
]
aggregator_config = AggregatorConfig("tests/data/aggregator_configs.json")
central_config = CentralConfig("tests/data/central_configs.json")


def make_node(path_learning, path_data):
    node = Node(learning_configs=path_learning, data=path_data)
    node.run(timeout=60)


def make_nodes(processes):

    nodes_data_config = [
        NodeDataConfig("tests/data/node_0/data_configs.json"),
        NodeDataConfig("tests/data/node_1/data_configs.json"),
        NodeDataConfig("tests/data/node_2/data_configs.json"),
    ]

    for learning, data in zip(nodes_learning_config, nodes_data_config):
        processes.append(Process(target=make_node, args=(learning, data)))
        processes[-1].start()
    sleep(3)
    print("")
    print("")


def make_server(processes):
    server = CentralServer(central_configs=central_config)
    processes.append(Process(target=server.run, args=(50,)))
    processes[-1].start()
    sleep(2)
    print("")
    print("")


def make_aggregator(processes):
    aggregator = Aggregator(aggregator_configs=aggregator_config)
    processes.append(Process(target=aggregator.run, args=(40,)))
    processes[-1].start()
    sleep(2)
    print("")
    print("")


def test_iris_server_aggr_node(clean):
    processes = []

    make_server(processes)
    make_aggregator(processes)
    make_nodes(processes)

    success = []
    while len(processes) > 0:
        for proc in processes:
            if proc.exitcode is not None:
                processes.remove(proc)
                success.append(proc.exitcode == 0)
    assert all(success)



import pytest

from ifra.decorator import emit


class SimpleEmitter:
    def __init__(self):
        self.messages = {"doing": None}

    def __getattr__(self, item):
        if item == "messages":
            raise ValueError("NodeMessenger object's 'messages' not set")
        if item == "path":
            raise ValueError("NodeMessenger object's 'path' not set")
        if item not in self.messages:
            raise ValueError(f"No messages named '{item}' was found")
        return self.messages[item]

    def __setattr__(self, item, value):
        if item == "messages" or item == "path":
            super().__setattr__(item, value)
            return
        if item not in self.messages and item not in self.__class__.DEFAULT_MESSAGES:
            raise ValueError(f"Message named '{item}' is not allowed")
        self.messages[item] = value

    def send(self, **kwargs):
        for key in kwargs:
            self.messages[key] = kwargs[key]


class Actor:
    def __init__(self):
        self.emitter = SimpleEmitter()

    @emit
    def func_1(self):
        assert self.emitter.doing == "func_1"

        @emit
        def func_2(self_):
            assert self.emitter.doing == "func_2"
            return "chat"

        @emit
        def func_3():
            pass

        @emit
        def func_4(a):
            a += 1
            pass

        chat = func_2(self)
        assert chat == "chat"

        assert self.emitter.doing == "func_1"

        with pytest.raises(ValueError):
            func_3()

        with pytest.raises(ValueError):
            func_4(0)

        return "chien"

    @emit
    def failing(self):
        raise ValueError("coucou")


def test_decorator():
    a = Actor()
    chien = a.func_1()
    assert chien == "chien"
    assert a.emitter.doing is None
    with pytest.raises(ValueError) as e:
        a.failing()
        assert a.emitter.error is not None
        assert "coucou" in a.emitter.error



import numpy as np
import pytest
from ruleskit import RuleSet, ClassificationRule, RegressionRule, HyperrectangleCondition, Activation
from ifra.diff_privacy import apply_diff_privacy

c_min = 0.10

np.random.seed(42)


# @pytest.mark.parametrize(
#     "ruleset, y, exp_preds",
#     [
#         (
#             RuleSet(
#                 [
#                     RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
#                 ],
#             ),
#             np.random.randint(0, 2, 100),
#             [0.09225888254859654]
#         )
#     ]
# )
# def test_privacy(ruleset, y, exp_preds):
#
#     for r, perd in zip(ruleset, exp_preds):
#         act = np.random.uniform(0, 1, 100)
#         act = act > (1 - c_min)
#         r._prediction = np.mean(np.extract(y, act))
#         r._coverage = sum(act) / len(act)
#         r._activation = Activation(act)
#         assert r._prediction == 0.05357142857142857
#
#     apply_diff_privacy(ruleset=ruleset, y=y, c_min=c_min)
#
#     for r, perd in zip(ruleset, exp_preds):
#         assert r._prediction == exp_preds
